type CellInputHandlers<T> = Record<string, (state: T, ...values: any) => T | Promise<T>>

type Tail<T extends Array<any>> =
	((...params: T) => any) extends ((param1: any, ...tail: infer U) => any) ? U : never

type CellOutputHandlers<T, S extends CellInputHandlers<T>> = {
	[key in keyof S]: (...values: Tail<Parameters<S[key]>>) =>
		ReturnType<S[key]> extends Promise<T> ? Promise<T> : T
}

interface Value<T> {
	value: T
}

interface SubscribeHandler {
	addOnChange: (callback: (field: string) => void, field: string) => void,
	removeOnChange: (callback: () => void) => void,
}

type Cell<T, S extends CellInputHandlers<T>> = CellOutputHandlers<T, S> & Value<T> & SubscribeHandler

function isPromise<T>(param: T | Promise<T>): param is Promise<T> {
	return typeof (<Promise<T>>param).then === 'function'
}

export function createCell<T, S extends CellInputHandlers<T>>(handlers: S, initialValue: T): () => Cell<T, S> {
	return function createCellConstructor() {
		let state = initialValue
		const onChangeListeners: Array<(field: string) => void> = []
		const onChangeFields: Array<string> = []

		const cellHandlers = <CellOutputHandlers<T, S>>Object.keys(handlers).reduce((result, name) => {
			const handler = handlers[name]

			result[name] = async (...values: Parameters<typeof handler>) => {
				const nextValue = handler.apply(null, [state, ...values])

				if (isPromise(nextValue)) {
					state = await nextValue
				} else {
					state = nextValue
				}

				onChangeListeners.forEach((handler, index) => {
					handler(onChangeFields[index])
				})

				return state
			}

			return result
		}, {})
		const cell = <CellOutputHandlers<T, S> & SubscribeHandler>{
			...cellHandlers,
			addOnChange(handler, field) {
				onChangeListeners.push(handler)
				onChangeFields.push(field)
			},
			removeOnChange(handler) {
				const index = onChangeListeners.indexOf(handler)

				if (index > -1) {
					onChangeListeners.splice(index, 1)
				}
			}
		}

		Object.defineProperty(cell, 'value', {
			get() {
				return state
			}
		})

		return cell as Cell<T, S>
	}
}
